import { autocompletion } from '@codemirror/autocomplete';
import { linter } from '@codemirror/lint';
import { Facet } from '@codemirror/state';
import { hoverTooltip, showTooltip } from '@codemirror/tooltip';
import { ViewPlugin, EditorView } from '@codemirror/view';
import { WebSocketTransport, RequestManager, Client } from '@open-rpc/client-js';

function _classCallCheck(instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
}

function _defineProperties(target, props) {
  for (var i = 0; i < props.length; i++) {
    var descriptor = props[i];
    descriptor.enumerable = descriptor.enumerable || false;
    descriptor.configurable = true;
    if ("value" in descriptor) descriptor.writable = true;
    Object.defineProperty(target, descriptor.key, descriptor);
  }
}

function _createClass(Constructor, protoProps, staticProps) {
  if (protoProps) _defineProperties(Constructor.prototype, protoProps);
  if (staticProps) _defineProperties(Constructor, staticProps);
  return Constructor;
}

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _unsupportedIterableToArray(arr, i) || _nonIterableRest();
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArrayLimit(arr, i) {
  if (typeof Symbol === "undefined" || !(Symbol.iterator in Object(arr))) return;
  var _arr = [];
  var _n = true;
  var _d = false;
  var _e = undefined;

  try {
    for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

  return arr2;
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}

var timeout = 10000;
var serverUri = Facet.define({
  combine: function combine(values) {
    return values.reduce(function (_, v) {
      return v;
    }, '');
  }
});
var rootUri = Facet.define({
  combine: function combine(values) {
    return values.reduce(function (_, v) {
      return v;
    }, '');
  }
});
var documentUri = Facet.define({
  combine: function combine(values) {
    return values.reduce(function (_, v) {
      return v;
    }, '');
  }
});
var languageId = Facet.define({
  combine: function combine(values) {
    return values.reduce(function (_, v) {
      return v;
    }, '');
  }
});

var LanguageServerPlugin = /*#__PURE__*/function () {
  function LanguageServerPlugin(view) {
    var _this = this;

    _classCallCheck(this, LanguageServerPlugin);

    this.view = view;
    this.rootUri = this.view.state.facet(rootUri);
    this.documentUri = this.view.state.facet(documentUri);
    this.languageId = this.view.state.facet(languageId);
    this.documentVersion = 0;
    this.promises = [];
    this.transport = new WebSocketTransport(this.view.state.facet(serverUri));
    this.requestManager = new RequestManager([this.transport]);
    this.client = new Client(this.requestManager);
    this.client.onNotification(function (data) {
      _this.processNotification(data);
    });
    this.initialize({
      documentText: this.view.state.doc.toString()
    });
  } // update({view, state, docChanged}) {
  // 	return null;
  // }


  _createClass(LanguageServerPlugin, [{
    key: "destroy",
    value: function destroy() {
      return this.client.close();
    }
  }, {
    key: "initialize",
    value: function initialize(_ref) {
      var _this2 = this;

      var documentText = _ref.documentText;
      this.client.request({
        method: 'initialize',
        params: {
          capabilities: {
            textDocument: {
              hover: {
                dynamicRegistration: true,
                contentFormat: ['plaintext', 'markdown']
              },
              synchronization: {
                dynamicRegistration: true,
                willSave: false,
                didSave: false,
                willSaveWaitUntil: false
              },
              completion: {
                dynamicRegistration: true,
                completionItem: {
                  snippetSupport: false,
                  commitCharactersSupport: true,
                  documentationFormat: ['plaintext', 'markdown'],
                  deprecatedSupport: false,
                  preselectSupport: false
                },
                contextSupport: false
              },
              signatureHelp: {
                dynamicRegistration: true,
                signatureInformation: {
                  documentationFormat: ['plaintext', 'markdown']
                }
              },
              declaration: {
                dynamicRegistration: true,
                linkSupport: true
              },
              definition: {
                dynamicRegistration: true,
                linkSupport: true
              },
              typeDefinition: {
                dynamicRegistration: true,
                linkSupport: true
              },
              implementation: {
                dynamicRegistration: true,
                linkSupport: true
              }
            },
            workspace: {
              didChangeConfiguration: {
                dynamicRegistration: true
              }
            }
          },
          initializationOptions: null,
          processId: null,
          rootUri: this.rootUri,
          workspaceFolders: [{
            name: 'root',
            uri: this.rootUri
          }]
        }
      }, timeout * 3).then(function (_ref2) {
        var capabilities = _ref2.capabilities;
            _ref2.serverInfo;
        _this2.capabilities = capabilities;

        _this2.client.notify({
          method: 'initialized',
          params: {}
        });

        _this2.client.notify({
          method: 'textDocument/didOpen',
          params: {
            textDocument: {
              uri: _this2.documentUri,
              languageId: _this2.languageId,
              text: documentText,
              version: _this2.documentVersion
            }
          }
        });

        _this2.ready = true;
      });
    }
  }, {
    key: "sendChange",
    value: function sendChange(_ref3) {
      var documentText = _ref3.documentText;

      if (!this.ready) {
        return;
      }

      try {
        this.client.notify({
          method: 'textDocument/didChange',
          params: {
            textDocument: {
              uri: this.documentUri,
              version: this.documentVersion++
            },
            contentChanges: [{
              text: documentText
            }]
          }
        });
      } catch (e) {
        console.error(e);
      }
    }
  }, {
    key: "requestDiagnostics",
    value: function requestDiagnostics(view) {
      var _this3 = this;

      this.sendChange({
        documentText: view.state.doc.toString()
      });
      return new Promise(function (_fulfill, _reject) {
        _this3.promises.push({
          type: 'diagnostics',
          fulfill: function fulfill() {
            return _fulfill.apply(void 0, arguments);
          },
          reject: function reject() {
            return _reject.apply(void 0, arguments);
          }
        });
      });
    }
  }, {
    key: "requestHoverTooltip",
    value: function requestHoverTooltip(view, _ref4) {
      var _this4 = this;

      var line = _ref4.line,
          character = _ref4.character;

      if (!this.ready || !this.capabilities.hoverProvider) {
        return null;
      }

      this.sendChange({
        documentText: view.state.doc.toString()
      });
      this.client.request({
        method: 'textDocument/hover',
        params: {
          textDocument: {
            uri: this.documentUri
          },
          position: {
            line: line,
            character: character
          }
        }
      }, timeout).then(function (result) {
        if (!result) {
          _this4.view.dispatch({
            reconfigure: {
              hoverTooltip: []
            }
          });

          return;
        }

        var contents = result.contents,
            range = result.range;
        var pos = posToOffset(view.state.doc, {
          line: line,
          character: character
        });
        var end;

        if (range) {
          pos = posToOffset(view.state.doc, range.start);
          end = posToOffset(view.state.doc, range.end);
        }

        _this4.view.dispatch({
          reconfigure: {
            hoverTooltip: showTooltip.of({
              pos: pos,
              end: end,
              create: function create(view) {
                var dom = document.createElement('div');
                dom.textContent = formatContents(contents);
                return {
                  dom: dom
                };
              },
              style: 'documentation',
              above: true
            })
          }
        });
      });
      return null;
    }
  }, {
    key: "requestCompletion",
    value: function requestCompletion(context, _ref5, _ref6) {
      var line = _ref5.line,
          character = _ref5.character;
      var triggerKind = _ref6.triggerKind,
          triggerCharacter = _ref6.triggerCharacter;

      if (!this.ready || !this.capabilities.completionProvider) {
        return null;
      }

      this.sendChange({
        documentText: context.state.doc.toString()
      });
      return this.client.request({
        method: 'textDocument/completion',
        params: {
          textDocument: {
            uri: this.documentUri
          },
          position: {
            line: line,
            character: character
          },
          context: {
            triggerKind: triggerKind,
            triggerCharacter: triggerCharacter
          }
        }
      }, timeout).then(function (result) {
        if (!result) {
          return null;
        }

        var options = result.items.map(function (_ref7) {
          var detail = _ref7.detail,
              label = _ref7.label,
              kind = _ref7.kind,
              textEdit = _ref7.textEdit,
              documentation = _ref7.documentation,
              sortText = _ref7.sortText,
              filterText = _ref7.filterText;
          var completion;
          completion = {
            label: label,
            detail: detail,
            apply: label,
            type: {
              1: 'text',
              2: 'method',
              3: 'function',
              // 4: Constructor
              // 5: Field
              6: 'variable',
              7: 'class',
              8: 'interface',
              // 9: Module
              10: 'property',
              // 11: Unit
              // 12: Value
              13: 'enum',
              14: 'keyword',
              // 15: Snippet
              // 16: Color
              // 17: File
              // 18: Reference
              // 19: Folder
              // 20: EnumMember
              21: 'constant'
            }[kind],
            sortText: label,
            filterText: label
          };

          if (textEdit) {
            completion.apply = textEdit.newText;
          }

          if (documentation) {
            completion.info = formatContents(documentation);
          }

          if (sortText) {
            completion.sortText = sortText;
          }

          if (filterText) {
            completion.filterText = filterText;
          }

          return completion;
        });

        var _prefixMatch = prefixMatch(options),
            _prefixMatch2 = _slicedToArray(_prefixMatch, 2);
            _prefixMatch2[0];
            var match = _prefixMatch2[1];

        var token = context.matchBefore(match);
        var pos = context.pos;

        if (token) {
          pos = token.from;
          var word = token.text.toLowerCase();

          if (/^\w+$/.test(word)) {
            options = options.filter(function (_ref8) {
              var filterText = _ref8.filterText;
              return filterText.toLowerCase().indexOf(word) === 0;
            });
            options = options.sort(function (a, b) {
              switch (true) {
                case a.apply.indexOf(token.text) === 0 && b.apply.indexOf(token.text) !== 0:
                  return -1;

                case a.apply.indexOf(token.text) !== 0 && b.apply.indexOf(token.text) === 0:
                  return 1;

                default:
                  return 0;
              }
            });
          }
        }

        return {
          from: pos,
          options: options
        };
      });
    }
  }, {
    key: "processNotification",
    value: function processNotification(_ref9) {
      var method = _ref9.method;
      var e;

      try {
        switch (method) {
          case 'textDocument/publishDiagnostics':
            this.processDiagnostics.apply(this, arguments);
        }
      } catch (error) {
        e = error;
        return console.error(e);
      }
    }
  }, {
    key: "processDiagnostics",
    value: function processDiagnostics(_ref10) {
      var _this5 = this;

      var params = _ref10.params;
      var annotations = params.diagnostics.map(function (_ref11) {
        var range = _ref11.range,
            message = _ref11.message,
            severity = _ref11.severity;
        return {
          from: posToOffset(_this5.view.state.doc, range.start),
          to: posToOffset(_this5.view.state.doc, range.end),
          severity: {
            1: 'error',
            2: 'warning',
            3: 'info',
            4: 'info'
          }[severity],
          message: message
        };
      });
      annotations.sort(function (a, b) {
        switch (true) {
          case a.from < b.from:
            return -1;

          case a.from > b.from:
            return 1;

          default:
            return 0;
        }
      });
      this.promises = this.promises.filter(function (p) {
        if (p.type === 'diagnostics') {
          p.fulfill(annotations);
          return false;
        } else {
          return true;
        }
      });
    }
  }, {
    key: "clearTooltip",
    value: function clearTooltip() {
      return this.view.dispatch({
        reconfigure: {
          hoverTooltip: []
        }
      });
    }
  }]);

  return LanguageServerPlugin;
}();
function languageServer(options) {
  var plugin;
  return [serverUri.of(options.serverUri), rootUri.of(options.rootUri), documentUri.of(options.documentUri), languageId.of(options.languageId), ViewPlugin.define(function (view) {
    plugin = new LanguageServerPlugin(view);
    return plugin;
  }, {
    eventHandlers: {
      keyup: function keyup() {
        return this.clearTooltip();
      },
      click: function click() {
        return this.clearTooltip();
      }
    }
  }), linter(function (view) {
    return plugin != null ? plugin.requestDiagnostics(view) : null;
  }), hoverTooltip(function (view, pos, side) {
    return plugin != null ? plugin.requestHoverTooltip(view, offsetToPos(view.state.doc, pos)) : null;
  }), autocompletion({
    override: [function (context) {
      if (plugin == null) {
        return null;
      }

      var state = context.state,
          pos = context.pos,
          explicit = context.explicit;
      var line = state.doc.lineAt(pos);
      var trigKind = 1; // Invoked

      var trigChar = void 0;

      if (!explicit && plugin.capabilities && plugin.capabilities.completionProvider && plugin.capabilities.completionProvider.triggerCharacters && plugin.capabilities.completionProvider.triggerCharacters.indexOf(line.text[pos - line.from - 1]) >= 0) {
        trigKind = 2; // TriggerCharacter

        trigChar = line.text[pos - line.from - 1];
      }

      if (trigKind === 1 && !context.matchBefore(/\w+$/)) {
        return null;
      }

      return plugin.requestCompletion(context, offsetToPos(state.doc, pos), {
        triggerKind: trigKind,
        triggerCharacter: trigChar
      });
    }]
  }), baseTheme];
}

function posToOffset(doc, pos) {
  return doc.line(pos.line + 1).from + pos.character;
}

function offsetToPos(doc, offset) {
  var line;
  line = doc.lineAt(offset);
  return {
    line: line.number - 1,
    character: offset - line.from
  };
}

function formatContents(contents) {
  var c, j, len, text;

  if (Array.isArray(contents)) {
    text = '';

    for (j = 0, len = contents.length; j < len; j++) {
      c = contents[j];
      text += formatContents(c) + '\n\n';
    }

    return text;
  } else if (typeof contents === 'string') {
    return contents;
  } else {
    return contents.value;
  }
}

function toSet(chars) {
  var flat, preamble, words;
  preamble = "";
  flat = Object.keys(chars).join('');
  words = /\w/.test(flat);

  if (words) {
    preamble += "\\w";
    flat = flat.replace(/\w/g, '');
  }

  return "[".concat(preamble).concat(flat.replace(/[^\w\s]/g, '\\$&'), "]");
}

function prefixMatch(options) {
  var apply, first, i, j, k, len, ref, rest, source;
  first = {};
  rest = {};

  for (j = 0, len = options.length; j < len; j++) {
    apply = options[j].apply;
    first[apply[0]] = true;

    for (i = k = 1, ref = apply.length; 1 <= ref ? k < ref : k > ref; i = 1 <= ref ? ++k : --k) {
      rest[apply[i]] = true;
    }
  }

  source = toSet(first) + toSet(rest) + "*$";
  return [new RegExp("^" + source), new RegExp(source)];
}
var baseTheme = EditorView.baseTheme({
  '$tooltip.documentation': {
    display: 'block',
    marginLeft: '0',
    padding: '3px 6px 3px 8px',
    borderLeft: '5px solid #999',
    whiteSpace: 'pre'
  },
  '$tooltip.lint': {
    whiteSpace: 'pre'
  }
});

export { languageServer };
